1000 ' SAVE "grape",A
1010 'lprint date$, time$: llist
1020 '
1030 ' Simple N-body program
1040 '
1050 ' Driver for GRAPE-1
1060 '
1070 ' J. Makino 89/09/13 (adopted from nbody.bas)
1080 '
1090 ' standard leap frog integrator is formulated as
1100 '
1110 ' x_i+1 = x_i + v_i+1/2 * dt
1120 ' v_i+1/2 = v_i-1/2 + a_i * dt
1130 ' where 
1140 ' a_i is the gravitational force obtained for position x_i
1150 '
1160 ' Here, we use the pseudo one-step formulation of the leap frog 
1170 ' 
1180 ' v_i+1/2 = v_i + a_i*dt/2
1190 ' x_i+1 = x_i + v_i+1/2 * dt
1200 ' v_i+1 = v_i+1/2 + a_i+1 * dt/2
1210 ' 
1220 ' This formula is easier to handle since velocity at t_i is at hand.
1230 '
1240 DEFSNG A-C,E-H,M,O-Z
1250 DEFINT I-L,N
1260 DEFDBL D
1270 INPUT "Number of particles"; N
1280 DIM X(N,3), V(N,3), A(N,3),M(N) ' allocate arrays for particle data
1290 GOSUB *INITIALIZE.GRAPE         ' initialize gpib and allocate work array
1300 GOSUB *INITIALIZE.SYSTEM        ' setup the uniform sphere dist.
1310 GOSUB *INITIALIZE.PARAMETERS    ' set parameters s.t. dt, tstep, eps
1320 GOSUB *STORE.CONSTANTS.TO.GRAPE ' store n & eps to grape
1330 GOSUB *INITIALIZE.PLOT          ' initialize screen
1340 GOSUB *PLOT.PARTICLES           ' plot initial condition
1350 GOSUB *PRINT.DIAG               ' print time, energy etc.
1360 ' setup the leap-frog integrator
1365 ' GOSUB *CALCULATE.ACCEL.ACCURATE : GOSUB *PRINT.ACCEL
1370 GOSUB *CALCULATE.ACCEL      '   : GOSUB *PRINT.ACCEL
1380 ' leap frog integrator loop
1390 WHILE T < TSTOP
1400    ' first calculate the velocity at the midpoint
1410    GOSUB *PUSH.VELOCITY.HALF
1420    ' then calculate the position and accelaration at the next time
1430    GOSUB *PUSH.POSITION.FULL
1440    ' GOSUB *CALCULATE.ACCEL.ACCURATE : GOSUB *PRINT.ACCEL
1445    GOSUB *CALCULATE.ACCEL   '      : GOSUB *PRINT.ACCEL
1450    ' then calculate the velocity at the next time
1460    GOSUB *PUSH.VELOCITY.HALF
1470    GOSUB *PLOT.PARTICLES
1480    T = T + DT  ' update time
1490    ' check if need to print out something
1500    IF T >= TNEXT THEN TNEXT = TNEXT + DTOUT : GOSUB *PRINT.DIAG
1510 WEND
1520 END
1530 '
1540 *PUSH.VELOCITY.HALF
1550    FOR I=1 TO N
1560       FOR K=1 TO 3
1570          V(I,K) = V(I,K) + .5*DT * A(I,K)
1580       NEXT K
1590 NEXT I
1600 RETURN
1610 '
1620 *PUSH.POSITION.FULL
1630 FOR I=1 TO N
1640    FOR K=1 TO 3
1650       X(I,K) = X(I,K) + DT * V(I,K)
1660    NEXT K
1670 NEXT I
1680 RETURN
1690 '
1700 *INITIALIZE.SYSTEM
1710 INPUT "random seed";ISEED
1720 MASS = 1!/N                                   ' set mass to 1/N
1730 FOR I = 1 TO N : M(I) = MASS : NEXT
1740 RANDOMIZE ISEED
1750 DIM WORK(3)
1760 FOR I = 1 TO N                                ' x and v is unform dist.
1770    GOSUB *GET.SPHERE                          ' within a unit sphere
1780    FOR K = 1 TO 3 : X(I,K) = WORK(K) : NEXT
1790    GOSUB *GET.SPHERE
1800    FOR K = 1 TO 3 : V(I,K) = WORK(K) : NEXT
1810 NEXT I
1820 DIM XOFF(3), VOFF(3)                          ' adjust x & v so that 
1830 MSUM = 0!                                     ' cm lies in the origin
1840 FOR I = 1 TO N
1850    FOR K = 1 TO 3
1860       XOFF(K) = XOFF(K) + M(I)*X(I,K)
1870       VOFF(K) = VOFF(K) + M(I)*V(I,K)
1880    NEXT K
1890   MSUM = MSUM + M(I)
1900 NEXT I
1910 FOR K = 1 TO 3 
1920    XOFF(K) = XOFF(K)/MSUM 
1930    VOFF(K) = VOFF(K)/MSUM 
1940 NEXT
1950 FOR I = 1 TO N : FOR K = 1 TO 3
1960    X(I,K) = X(I,K) - XOFF(K)
1970    V(I,K) = V(I,K) - VOFF(K)
1980 NEXT K : NEXT I
1990 RETURN
2000 '
2010 *GET.SPHERE                                   ' obtain random point in
2020 FOR K = 1 TO 3                                ' a unit sphere
2030    WORK(K) = RND(1) *2! -1!
2040 NEXT K
2050 IF WORK(1)^2+WORK(2)^2+WORK(3)^2 > 1! THEN *GET.SPHERE
2060 RETURN
2070 '
2080 *INITIALIZE.PARAMETERS
2090 INPUT "dt, tstop, dtout, eps"; DT, TSTOP, DTOUT, EPS
2100 T = 0!
2110 TNEXT = DTOUT
2120 EPS2 = EPS*EPS
2130 PRINT USING "N = ######"; N;
2140 PRINT USING "  dt=#.###^^^^"; DT;
2150 PRINT USING "  tstop=#.###^^^^"; TSTOP;
2160 PRINT USING "  dtout=#.###^^^^"; DTOUT;
2170 PRINT USING "  eps=#.###^^^^"; EPS;
2180 PRINT
2190 RETURN
2200 '
2210 *PRINT.DIAG
2220 'print total energy, virial ratio, c.m. motion and angular momentum
2230 DIM CM(3), CMV(3), AM(3)
2240 ' calculate potential energy
2250 PE = 0!
2260 FOR I = 1 TO N
2270    FOR J = I+1 TO N
2280       RIJ = (X(I,1)-X(J,1))^2+(X(I,2)-X(J,2))^2+(X(I,3)-X(J,3))^2+EPS2
2290       PE = PE - M(I)*M(J)/SQR(RIJ)
2300    NEXT
2310 NEXT
2320 'calculate kinetic energy and c.m. motion, angular momentum
2330 EK = 0  : FOR K = 1 TO 3 : CM(K) = 0 : CMV(K) = 0 : AM(K) = 0:NEXT
2340 MSUM = 0
2350 FOR I = 1 TO N
2360    EK = EK + .5*M(I)*(V(I,1)^2+V(I,2)^2+V(I,3)^2)
2370    FOR K = 1 TO 3
2380       CM(K) = CM(K) + M(I)*X(I,K)
2390       CMV(K) = CMV(K) + M(I)*V(I,K)
2400       K1 = 1 + (K MOD 3)
2410       K2 = 1 + ((K+1) MOD 3)
2420       AM(K) = AM(K) + M(I)*(X(I,K1)*V(I,K2)-X(I,K2)*V(I,K1))
2430    NEXT K
2440    MSUM = MSUM + M(I)
2450 NEXT
2460 ' scale cm motion by total mass
2470 FOR K = 1 TO 3
2480    CM(K) = CM(K)/MSUM
2490    CMV(K) = CMV(K)/MSUM
2500 NEXT K
2510 PRINT USING "T=#####.###"; T ;
2520 PRINT USING " E=#.#######"; EK+PE ;
2530 PRINT USING " V.R.=#.####"; -EK/PE;
2540 PRINT
2550 PRINT "CM :";:FOR K=1 TO 3 : PRINT USING " #.######^^^^";CM(K);:NEXT
2560 PRINT
2570 PRINT "CMv:";:FOR K=1 TO 3 : PRINT USING " #.######^^^^";CMV(K);:NEXT
2580 PRINT
2590 PRINT "am :";:FOR K=1 TO 3 : PRINT USING " #.######^^^^"; AM(K);:NEXT
2600 PRINT
2610 PRINT DATE$+" "+TIME$
2620 PRINT
2630 ERASE CM, CMV, AM
2640 RETURN
2650 '
2660 *CALCULATE.ACCEL
2670 ' first clear the acceleration array
2680 FOR I = 1 TO N
2690    FOR K = 1 TO 3
2700       A(I,K) = 0!
2710    NEXT K
2720 NEXT I
2730 ' send xjs to grape 
2740 FOR IPART% = 1 TO N
2750    FOR J=1 TO 3 : XDATA!(J) = X(IPART%, J) : NEXT
2760    IJMODE% = 1 : GOSUB *STORE.XIJ.DATA
2770 NEXT IPART%
2780 ' calculate force on each particle
2790 FOR IX = 1 TO N   '  : PRINT "calclate accel : ix =", IX
2800    FOR J=1 TO 3 : XDATA!(J) = X(IX, J) : NEXT
2810    IJMODE% = 0 : GOSUB *STORE.XIJ.DATA
2820    GOSUB *SET.GRAPE.TALKER
2830    GOSUB *GET.RESULT
2840    FOR J=1 TO 3 : A(IX,J) = ACCEL#(J)*M(1) : NEXT
2850    GOSUB *SET.GRAPE.LISTENER
2860 NEXT IX
2870 RETURN
2880 '
2890 *SAVE.FILE
2900 'SAVE "nbody",A
2910 END
2920 *PLOT.PARTICLES
2930 FOR I = 1 TO N
2940    PSET (X(I,1),X(I,2)),(I MOD 7)+1
2950 NEXT
2960 RETURN
2970 '
2980 *INITIALIZE.PLOT
2990 SCREEN 3
3000 INPUT "enter max y size"; WY
3010 WX1 = 2.2*WY
3020 WINDOW(-WX1,-WY)-(WY,WY)
3030 VIEW (0,0)-(639,399)
3040 CLS 2
3050 RETURN
3060 *INITIALIZE.GPIB
3070 ISET IFC
3080 PRINT "IFC end "
3090 '
3100 ISET REN
3110 PRINT "REN end"
3120 RETURN
3130 *SET.GRAPE.LISTENER
3140 WBYTE &H4C; ' address myself as the talker
3150  ' PRINT "TALKER command end"
3160 WBYTE &H20; ' address GRAPE-1 as the listener
3170 ' PRINT "LISTENER command end"
3180 RETURN
3190 *SET.GRAPE.TALKER
3200 WBYTE &H40; ' address GRAPE-1 as the talker
3210 ' PRINT "TALKER command end"
3220 WBYTE &H2C; ' address myself as the listener
3230 ' PRINT "LISTENER command end"
3240 RETURN
3250 '
3260 *GET.BYTES.FROM.INTEGER
3270 ' iget.bytes.input% : integer   input    input data
3280 ' iget.bytes.lobyte%: integer   output   low byte of input data
3290 ' iget.bytes.hibyte%: integer   output   high byte of input data
3300 '
3310 D$=MKI$(IGET.BYTES.INPUT%) : ' PRINT "length of d$=",LEN(D$)
3320 IGET.BYTES.LOBYTE% = ASC(LEFT$(D$,1))
3330 IGET.BYTES.HIBYTE% = ASC(RIGHT$(D$,1))
3340 ' PRINT "*get.bytes.from.integer", HEX$(IGET.BYTES.INPUT%)," ";
3350 ' PRINT HEX$(IGET.BYTES.HIBYTE%)," ",HEX$(IGET.BYTES.LOBYTE%)
3360 ' PRINT IGET.BYTES.INPUT%, IGET.BYTES.LOBYTE%, IGET.BYTES.HIBYTE%
3370 RETURN
3380 '
3390 *SEND.XIJ.DATA
3400 '
3410 ' ipart : integer     address of the particle
3420 ' IXJ%(3) :integer     particle position data
3430 ' ijmode% :integer     0 : register write, 1 : memory write
3440 '
3450 IGET.BYTES.INPUT% = IPART% : GOSUB *GET.BYTES.FROM.INTEGER
3460 ' PRINT "*send.xij.data", IXJ%(1),IXJ%(2),IXJ%(3) 
3470 IP.HI% = IGET.BYTES.HIBYTE% : IP.LO% = IGET.BYTES.LOBYTE%
3480 FOR I%=1 TO 3
3490    D$=MKI$(IXJ%(I%))
3500    LOBYTE% = ASC(LEFT$(D$,1))
3510    HIBYTE% = ASC(RIGHT$(D$,1))
3520    IADDRESS% = (I%-1) * &H20
3530    IF IJMODE% = 0 THEN IADDRESS% = IADDRESS% + &HA0
3540    WBYTE ; IADDRESS%,IP.HI%,IP.LO%,HIBYTE%,LOBYTE%
3550    ' PRINT   IADDRESS%,IP.HI%,IP.LO%,HIBYTE%,LOBYTE%
3560 NEXT
3570 RETURN
3580 '
3590 *SEND.XI.DATA
3600 ' argument list
3610 '
3620 ' IXJ%(3) :integer     particle position data
3630 '
3640 FOR I=1 TO 3
3650    D$=MKI$(IXJ%(I))
3660    HIBYTE% = ASC(LEFT$(D$,1))
3670    LOBYTE% = ASC(RIGHT$(D$,1))
3680    IADDRESS% = (I-1) * &H20 + &HA0
3690    WBYTE ; IADDRESS%,0,0,HIBYTE%,LOBYTE%
3700 NEXT
3710 RETURN
3720 '
3730 *SEND.EPSILON
3740 '
3750 ' ieps%   integer input     8bit float format data for epsilon
3760 '
3770 WBYTE ;&H60,0,0,IEPS%,IEPS% ' store 0 at epsilon
3780 RETURN
3790 '
3800 *SEND.PARTICLE.NUMBER
3810 '
3820 ' n%      integer  input    particle number
3830 PRINT "send particle number ", N%
3840  IGET.BYTES.INPUT% = N%+2 : GOSUB *GET.BYTES.FROM.INTEGER
3850 WBYTE ;&H80,0,0,IGET.BYTES.HIBYTE%,IGET.BYTES.LOBYTE% ' store n plus 2
3860 RETURN
3870 '
3940 *GET.RESULT
3950 '
3960 ' accel#(3)   double   output   force calculated by GRAPE-1
3970 ' 
3980 FORCEMAX# = 2#^47
3990 FORCEOFFSET# = 2#^48
4000 FORCESCALE# = 2#^(-18)
4010 FOR IWORK%=1 TO 18 : RBYTE; RESULT%(IWORK%)  : NEXT
4020 'FOR IWORK%=1 TO 18 : PRINT RESULT%(IWORK%); : NEXT: PRINT
4030 'PRINT " nomal return from read data from GRAPE"
4040 FOR IWORK% = 1 TO 3 
4050    ACCEL#(IWORK%) = 0
4060    FOR JWORK% = (IWORK%-1)*6+1 TO IWORK%*6
4070       ACCEL#(IWORK%) = ACCEL#(IWORK%) * 256# + RESULT%(JWORK%)
4080    NEXT JWORK%
4090    ' PRINT "get.result", IWORK%, ACCEL#(IWORK%)
4100    IF ACCEL#(IWORK%) < FORCEMAX# THEN *FORCE.OK
4110      ACCEL#(IWORK%) = ACCEL#(IWORK%) - FORCEOFFSET#
4120    *FORCE.OK
4130    ACCEL#(IWORK%) = ACCEL#(IWORK%) *FORCESCALE#
4140    ' PRINT "get.result", IWORK%, ACCEL#(IWORK%)
4150 NEXT
4160 RETURN
4170 '
4180 *STORE.XIJ.DATA
4190 '
4200 ' ipart% integer input identity of the particle
4210 ' xdata!(3) real input position data
4220 ' ijmode%    integer input  0 : store xi(fixed one), ipart% is ignored 
4230 '                          1 : store xj(loop over), ipart% is used
4240 ' 
4250 FOR IWORK% = 1 TO 3
4260    XDUMMY! = XDATA!(IWORK%) * 128
4270    IF XDUMMY! < -32767! THEN XDUMMY! = -32767!
4280    IF XDUMMY! > 32767! THEN XDUMMY! = 32767!
4290    IXJ%(IWORK%) = XDUMMY!
4300    ' PRINT "store.xij.data, j, x, ix", IWORK%, XDATA!(IWORK%), IXJ%(IWORK%)
4310 NEXT
4320 GOSUB *SEND.XIJ.DATA
4330 RETURN
4340 '
4350 *INITIALIZE.GRAPE
4360 DIM RESULT%(18)
4370 DIM IXJ%(3)
4380 DIM XDATA!(3)
4390 DIM ACCEL#(3)
4400 PRINT "start initializing the GPIB interface"
4410 GOSUB *INITIALIZE.GPIB
4420 GOSUB *SET.GRAPE.LISTENER
4430 RETURN
4440 '
4450 *STORE.CONSTANTS.TO.GRAPE
4460 SRFACT = 1!/128 : FSSCAL = (256 * .9)/2!/LOG(32768!)
4470 IF EPS > 0 THEN IEPS% = LOG(EPS*128)*FSSCAL + 1 ELSE IEPS% = 0
4475 PRINT "eps : ", EPS, IEPS%
4480 GOSUB *SEND.EPSILON
4481 N% = N : GOSUB *SEND.PARTICLE.NUMBER
4490 RETURN
4500 '
4510 XDATA!(1) = 0 : XDATA!(2) = 0 : XDATA!(3) = 0 : IJMODE% = 1
4520 INPUT "xj"; XDATA!(1),XDATA!(2),XDATA!(3)
4530 N%=N: GOSUB *SEND.PARTICLE.NUMBER : PRINT N, N%
4540 FOR IPART% = 1 TO N%: GOSUB *STORE.XIJ.DATA :NEXT
4550 INPUT "j, xj"; IPART%, XDATA!(1),XDATA!(2),XDATA!(3)
4560 GOSUB *STORE.XIJ.DATA
4570 FOR II = 3 TO 5
4580    INPUT "xi"; XDATA!(1),XDATA!(2),XDATA!(3)
4590    IJMODE% = 0 : GOSUB *STORE.XIJ.DATA
4600    PRINT "all data should have been sent ----"
4610    GOSUB *SET.GRAPE.TALKER
4620    GOSUB *GET.RESULT
4630    GOSUB *SET.GRAPE.LISTENER
4640 NEXT
20000 *CALCULATE.ACCEL.ACCURATE
20010 ' first clear the acceleration array
20020 FOR I = 1 TO N
20030    FOR K = 1 TO 3
20040       A(I,K) = 0!
20050    NEXT K
20060 NEXT I
20070 ' calculte N(N-1)/2 interactions directly
20080 FOR I = 1 TO N-1
20090    FOR J= I+1 TO N
20100       DX = X(I,1) - X(J,1)
20110       DY = X(I,2) - X(J,2)
20120       DZ = X(I,3) - X(J,3)
20130       RIJ2 = DX*DX + DY*DY + DZ*DZ + EPS2
20140       RFACT = 1!/(RIJ2*SQR(RIJ2))
20150       AIFACT = -M(J)*RFACT
20160       AJFACT = M(I)*RFACT
20170       A(I,1) = A(I,1) + DX*AIFACT
20180       A(I,2) = A(I,2) + DY*AIFACT
20190       A(I,3) = A(I,3) + DZ*AIFACT
20200       A(J,1) = A(J,1) + DX*AJFACT
20210       A(J,2) = A(J,2) + DY*AJFACT
20220       A(J,3) = A(J,3) + DZ*AJFACT
20230    NEXT J
20240 NEXT I
20250 RETURN
20260 '
30000 *PRINT.ACCEL
30005 PRINT " calculated gravity "
30010 FOR I = 1 TO N
30020    PRINT I, A(I,1), A(I,2), A(I,3)
30030 NEXT
30040 RETURN
