//
// checkforsoftbhb.C
//
// Version 2001/4/23 Jun Makino
// Created ---
// check for soft binary of bh and field star
// and adjust the velocity of the field star if necessary;
//
// 


#include "pit_system.h"
#include  <fstream.h>

real check_and_adjust_pair(particle * pi, particle * pj, real ecut)
{
    vector r = pj->get_pos() - pi->get_pos();
    vector v = pj->get_vel() - pi->get_vel();
    real msum= pi->get_mass()+pj->get_mass();
    real eb = 0.5*(v*v)-msum/sqrt(r*r);
    real eorig = eb;
    if (eb < ecut){
	cerr << "Bound pair" << pi->get_index() << " " << pj->get_index()
	     << " eb = " << eb << endl;
	cerr << "R = " << pj->get_pos() << " " <<  pi->get_pos()<<endl;
	cerr << "V = " << pj->get_vel() << " " <<  pi->get_vel()<<endl;
	real de = ecut/2-eb;
	real oldv2 = v*v;
	real newv2 = oldv2 + de*2;
	cerr << "v2 change " << oldv2 <<"->"<<newv2 <<endl;
	vector dv = v/sqrt(oldv2)*(sqrt(newv2)-sqrt(oldv2));
	cerr << "v change " << dv <<endl;
	vector oldv = pj->get_vel();
	pj->set_vel(oldv+dv);
	cerr << "v change "  << " " <<oldv << " => " <<pj->get_vel()<<endl;
	r = pj->get_pos() - pi->get_pos();
	v = pj->get_vel() - pi->get_vel();
	eb = 0.5*(v*v)-msum/sqrt(r*r);
	cerr << "New eb = " <<eb <<endl;
	
    }
    return eorig;
}
void adjust_cmv(particle * pb,
		int nbody)
    
{
    real msum = 0;
    vector vsum = 0;
    for(int i=0;i<nbody;i++){
	real mass = pb[i].get_mass();
	msum += mass;
	vsum += mass*pb[i].get_vel();
    }
    vsum /= msum;
    cerr << "Before adjust msum=" << msum << " cmv = "<<vsum<<endl;
    for(int i=0;i<nbody;i++){
	pb[i].set_vel(pb[i].get_vel()-vsum);
    }
}    


void check_for_soft_bh(particle * pb,
		       int nbody,
		       int nbh,
		       real ecut)
{

    int i,j;
    real ebmax = 0;
    for(i=0;i<nbh;i++){
	cerr << "Checking for bh = " << i <<endl;
	for(j=nbh;j<nbody;j++){
	    real eb = check_and_adjust_pair(pb+i,pb+j, ecut);
	    if(eb < ebmax)  ebmax = eb;
	}
    }
    cerr << "EBmax = "<< ebmax <<endl;
    adjust_cmv(pb, nbody);
}

void print_usage_and_exit()
{
    cerr <<
	 "usage: checksoftbhb -b # -e  #"
	 << "    for b number of BHs,\n"
	 << "        e cutoff energy per reduced mass\n";
    exit(1);
}

main(int argc, char **argv)
{
    particle_system s;			// input system

    extern char *poptarg;
    int pgetopt(int, char **, char *), c;
    bool b_flag = false;
    bool e_flag = false;
    int nbh = 0;
    real ecut = 0; 
    
    while ((c = pgetopt(argc, argv, "b:e:")) != -1) {
	switch (c) {
	    case 'b':
		b_flag = true;
		nbh = atoi(poptarg);
		break;
	    case 'e':
		e_flag = true;
		ecut = atof(poptarg);
		break;
	    case '?':
		print_usage_and_exit();
	}
    }
    cerr << "NBH = " << nbh <<endl; 
    cerr << "Ecut = " << ecut <<endl;

    s.read(cin);
    
    check_for_soft_bh(s.get_pb(), s.get_n(), nbh, ecut);
    cout.precision(HIGH_PRECISION);
    int ntotal = s.get_n();
    s.write(cout);
}
